# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os
import warnings
from collections import OrderedDict
from pathlib import Path

import numpy as np
import torch

from monai.config.type_definitions import PathLike
from monai.utils import ImageMetaKey as Key


class CSVSaver:
    """
    Save the data in a dictionary format cache, and write to a CSV file finally.
    Typically, the data can be classification predictions, call `save` for single data
    or call `save_batch` to save a batch of data together, and call `finalize` to write
    the cached data into CSV file. If no metadata provided, use index from 0 to save data.
    Note that this saver can't support multi-processing because it reads / writes single
    CSV file and can't guarantee the data order in multi-processing situation.

    """

    def __init__(
        self,
        output_dir: PathLike = "./",
        filename: str = "predictions.csv",
        overwrite: bool = True,
        flush: bool = False,
        delimiter: str = ",",
    ) -> None:
        """
        Args:
            output_dir: output CSV file directory.
            filename: name of the saved CSV file name.
            overwrite: whether to overwriting existing CSV file content, if True, will clear the file before saving.
                otherwise, will append new content to the CSV file.
            flush: whether to write the cache data to CSV file immediately when `save_batch` and clear the cache.
                default to False.
            delimiter: the delimiter character in the saved file, default to "," as the default output type is `csv`.
                to be consistent with: https://docs.python.org/3/library/csv.html#csv.Dialect.delimiter.

        """
        self.output_dir = Path(output_dir)
        self._cache_dict: OrderedDict = OrderedDict()
        if not (isinstance(filename, str) and filename[-4:] == ".csv"):
            warnings.warn("CSV filename is not a string ends with '.csv'.")
        self._filepath = self.output_dir / filename
        if self._filepath.exists() and overwrite:
            os.remove(self._filepath)

        self.flush = flush
        self.delimiter = delimiter
        self._data_index = 0

    def finalize(self) -> None:
        """
        Writes the cached dict to a csv

        """
        if not self.output_dir.exists():
            self.output_dir.mkdir(parents=True, exist_ok=True)
        with open(self._filepath, "a") as f:
            for k, v in self._cache_dict.items():
                f.write(k)
                for result in v.flatten():
                    f.write(self.delimiter + str(result))
                f.write("\n")
        # clear cache content after writing
        self.reset_cache()

    def save(self, data: torch.Tensor | np.ndarray, meta_data: dict | None = None) -> None:
        """Save data into the cache dictionary. The metadata should have the following key:
            - ``'filename_or_obj'`` -- save the data corresponding to file name or object.
        If meta_data is None, use the default index from 0 to save data instead.

        Args:
            data: target data content that save into cache.
            meta_data: the metadata information corresponding to the data.

        """
        save_key = meta_data[Key.FILENAME_OR_OBJ] if meta_data else str(self._data_index)
        self._data_index += 1
        if isinstance(data, torch.Tensor):
            data = data.detach().cpu().numpy()
        self._cache_dict[save_key] = np.asarray(data, dtype=float)

    def save_batch(self, batch_data: torch.Tensor | np.ndarray, meta_data: dict | None = None) -> None:
        """Save a batch of data into the cache dictionary.

        Args:
            batch_data: target batch data content that save into cache.
            meta_data: every key-value in the meta_data is corresponding to 1 batch of data.

        """
        for i, data in enumerate(batch_data):  # save a batch of files
            self.save(data, {k: meta_data[k][i] for k in meta_data} if meta_data else None)

        if self.flush:
            self.finalize()

    def get_cache(self) -> OrderedDict:
        """Get the cache dictionary, key is filename and value is the corresponding data"""

        return self._cache_dict

    def reset_cache(self) -> None:
        """Clear the cache dictionary content"""
        self._cache_dict.clear()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import torch
from ignite.engine import Engine

from monai.data import Dataset
from monai.engines import Evaluator
from monai.handlers import ValidationHandler


class TestEvaluator(Evaluator):
    __test__ = False  # indicate to pytest that this class is not intended for collection

    def _iteration(self, engine, batchdata):
        engine.state.output = "called"
        return engine.state.output


class TestHandlerValidation(unittest.TestCase):

    def test_content(self):
        data = [0] * 8

        # set up engine
        def _train_func(engine, batch):
            pass

        engine = Engine(_train_func)

        # set up testing handler
        val_data_loader = torch.utils.data.DataLoader(Dataset(data))
        evaluator = TestEvaluator(torch.device("cpu:0"), val_data_loader)
        ValidationHandler(interval=2, validator=evaluator, exec_at_start=True).attach(engine)
        # test execution at start
        engine.run(data, max_epochs=1)
        self.assertEqual(evaluator.state.max_epochs, 1)
        self.assertEqual(evaluator.state.epoch_length, 8)
        self.assertEqual(evaluator.state.output, "called")

        engine.run(data, max_epochs=5)
        self.assertEqual(evaluator.state.max_epochs, 4)
        self.assertEqual(evaluator.state.epoch_length, 8)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest
from typing import Any

import torch
from parameterized import parameterized

from monai.networks.nets import DenseNet, DenseNet121
from monai.visualize import OcclusionSensitivity


class DenseNetAdjoint(DenseNet121):

    def __call__(self, x, adjoint_info):
        if adjoint_info != 42:
            raise ValueError
        return super().__call__(x)


device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
out_channels_2d = 4
out_channels_3d = 3
model_2d = DenseNet121(spatial_dims=2, in_channels=1, out_channels=out_channels_2d).to(device)
model_2d_2c = DenseNet121(spatial_dims=2, in_channels=2, out_channels=out_channels_2d).to(device)
model_3d = DenseNet(
    spatial_dims=3, in_channels=1, out_channels=out_channels_3d, init_features=2, growth_rate=2, block_config=(6,)
).to(device)
model_2d_adjoint = DenseNetAdjoint(spatial_dims=2, in_channels=1, out_channels=out_channels_2d).to(device)
model_2d.eval()
model_2d_2c.eval()
model_3d.eval()
model_2d_adjoint.eval()

TESTS: list[Any] = []
TESTS_FAIL: list[Any] = []

# 2D w/ bounding box with all modes
for mode in ("gaussian", "mean_patch", "mean_img"):
    TESTS.append(
        [
            {"nn_module": model_2d, "mode": mode},
            {"x": torch.rand(1, 1, 48, 64).to(device), "b_box": [2, 40, 1, 62]},
            (1, out_channels_2d, 38, 61),
            (1, 1, 38, 61),
        ]
    )
# 3D w/ bounding box
TESTS.append(
    [
        {"nn_module": model_3d, "n_batch": 10, "mask_size": (16, 15, 14)},
        {"x": torch.rand(1, 1, 64, 32, 16).to(device), "b_box": [2, 43, -1, -1, -1, -1]},
        (1, out_channels_3d, 41, 32, 16),
        (1, 1, 41, 32, 16),
    ]
)
TESTS.append(
    [
        {"nn_module": model_3d, "n_batch": 10},
        {"x": torch.rand(1, 1, 6, 7, 8).to(device), "b_box": [1, 3, -1, -1, -1, -1]},
        (1, out_channels_3d, 2, 7, 8),
        (1, 1, 2, 7, 8),
    ]
)
TESTS.append(
    [
        {"nn_module": model_2d_2c},
        {"x": torch.rand(1, 2, 48, 64).to(device)},
        (1, out_channels_2d, 48, 64),
        (1, 1, 48, 64),
    ]
)
# 2D w/ bounding box and adjoint
TESTS.append(
    [
        {"nn_module": model_2d_adjoint},
        {"x": torch.rand(1, 1, 48, 64).to(device), "b_box": [2, 40, 1, 62], "adjoint_info": 42},
        (1, out_channels_2d, 38, 61),
        (1, 1, 38, 61),
    ]
)
# 2D should fail: bbox makes image too small
TESTS_FAIL.append(
    [{"nn_module": model_2d, "n_batch": 10, "mask_size": 200}, {"x": torch.rand(1, 1, 48, 64).to(device)}, ValueError]
)
# 2D should fail: batch > 1
TESTS_FAIL.append(
    [{"nn_module": model_2d, "n_batch": 10, "mask_size": 100}, {"x": torch.rand(2, 1, 48, 64).to(device)}, ValueError]
)
# 2D should fail: unknown mode
TESTS_FAIL.append(
    [{"nn_module": model_2d, "mode": "test"}, {"x": torch.rand(1, 1, 48, 64).to(device)}, NotImplementedError]
)


class TestComputeOcclusionSensitivity(unittest.TestCase):

    @parameterized.expand(TESTS)
    def test_shape(self, init_data, call_data, map_expected_shape, most_prob_expected_shape):
        occ_sens = OcclusionSensitivity(**init_data)
        m, most_prob = occ_sens(**call_data)
        self.assertTupleEqual(m.shape, map_expected_shape)
        self.assertTupleEqual(most_prob.shape, most_prob_expected_shape)
        # most probable class should be of type int, and should have min>=0, max<num_classes
        self.assertEqual(most_prob.dtype, torch.int64)
        self.assertGreaterEqual(most_prob.min(), 0)
        self.assertLess(most_prob.max(), m.shape[-1])

    @parameterized.expand(TESTS_FAIL)
    def test_fail(self, init_data, call_data, error_type):
        with self.assertRaises(error_type):
            occ_sens = OcclusionSensitivity(**init_data)
            occ_sens(**call_data)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from collections.abc import Callable

from monai.handlers.ignite_metric import IgniteMetricHandler
from monai.metrics import AveragePrecisionMetric
from monai.utils import Average


class AveragePrecision(IgniteMetricHandler):
    """
    Computes Average Precision (AP).
    accumulating predictions and the ground-truth during an epoch and applying `compute_average_precision`.

    Args:
        average: {``"macro"``, ``"weighted"``, ``"micro"``, ``"none"``}
            Type of averaging performed if not binary classification. Defaults to ``"macro"``.

            - ``"macro"``: calculate metrics for each label, and find their unweighted mean.
                This does not take label imbalance into account.
            - ``"weighted"``: calculate metrics for each label, and find their average,
                weighted by support (the number of true instances for each label).
            - ``"micro"``: calculate metrics globally by considering each element of the label
                indicator matrix as a label.
            - ``"none"``: the scores for each class are returned.

        output_transform: callable to extract `y_pred` and `y` from `ignite.engine.state.output` then
            construct `(y_pred, y)` pair, where `y_pred` and `y` can be `batch-first` Tensors or
            lists of `channel-first` Tensors. the form of `(y_pred, y)` is required by the `update()`.
            `engine.state` and `output_transform` inherit from the ignite concept:
            https://pytorch.org/ignite/concepts.html#state, explanation and usage example are in the tutorial:
            https://github.com/Project-MONAI/tutorials/blob/master/modules/batch_output_transform.ipynb.

    Note:
        Average Precision expects y to be comprised of 0's and 1's.
        y_pred must either be probability estimates or confidence values.

    """

    def __init__(self, average: Average | str = Average.MACRO, output_transform: Callable = lambda x: x) -> None:
        metric_fn = AveragePrecisionMetric(average=Average(average))
        super().__init__(metric_fn=metric_fn, output_transform=output_transform, save_details=False)



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import torch
from parameterized import parameterized

from monai.networks import eval_mode
from monai.networks.nets.unetr import UNETR
from tests.test_utils import SkipIfBeforePyTorchVersion, skip_if_quick, test_script_save

TEST_CASE_UNETR = []
for dropout_rate in [0.4]:
    for in_channels in [1]:
        for out_channels in [2]:
            for hidden_size in [768]:
                for img_size in [96, 128]:
                    for feature_size in [16]:
                        for num_heads in [8]:
                            for mlp_dim in [3072]:
                                for norm_name in ["instance"]:
                                    for proj_type in ["perceptron"]:
                                        for nd in (2, 3):
                                            test_case = [
                                                {
                                                    "in_channels": in_channels,
                                                    "out_channels": out_channels,
                                                    "img_size": (img_size,) * nd,
                                                    "hidden_size": hidden_size,
                                                    "feature_size": feature_size,
                                                    "norm_name": norm_name,
                                                    "mlp_dim": mlp_dim,
                                                    "num_heads": num_heads,
                                                    "proj_type": proj_type,
                                                    "dropout_rate": dropout_rate,
                                                    "conv_block": True,
                                                    "res_block": False,
                                                },
                                                (2, in_channels, *([img_size] * nd)),
                                                (2, out_channels, *([img_size] * nd)),
                                            ]
                                            if nd == 2:
                                                test_case[0]["spatial_dims"] = 2  # type: ignore
                                            TEST_CASE_UNETR.append(test_case)


@skip_if_quick
class TestUNETR(unittest.TestCase):
    @parameterized.expand(TEST_CASE_UNETR)
    def test_shape(self, input_param, input_shape, expected_shape):
        net = UNETR(**input_param)
        with eval_mode(net):
            result = net(torch.randn(input_shape))
            self.assertEqual(result.shape, expected_shape)

    def test_ill_arg(self):
        with self.assertRaises(ValueError):
            UNETR(
                in_channels=1,
                out_channels=3,
                img_size=(128, 128, 128),
                feature_size=16,
                hidden_size=128,
                mlp_dim=3072,
                num_heads=12,
                proj_type="conv",
                norm_name="instance",
                dropout_rate=5.0,
            )

        with self.assertRaises(ValueError):
            UNETR(
                in_channels=1,
                out_channels=4,
                img_size=(32, 32, 32),
                feature_size=32,
                hidden_size=512,
                mlp_dim=3072,
                num_heads=12,
                proj_type="conv",
                norm_name="instance",
                dropout_rate=0.5,
            )

        with self.assertRaises(ValueError):
            UNETR(
                in_channels=1,
                out_channels=3,
                img_size=(96, 96, 96),
                feature_size=16,
                hidden_size=512,
                mlp_dim=3072,
                num_heads=14,
                proj_type="conv",
                norm_name="batch",
                dropout_rate=0.4,
            )

        with self.assertRaises(ValueError):
            UNETR(
                in_channels=1,
                out_channels=4,
                img_size=(96, 96, 96),
                feature_size=8,
                hidden_size=768,
                mlp_dim=3072,
                num_heads=12,
                proj_type="perc",
                norm_name="instance",
                dropout_rate=0.2,
            )

    @parameterized.expand(TEST_CASE_UNETR)
    @SkipIfBeforePyTorchVersion((2, 0))
    def test_script(self, input_param, input_shape, _):
        net = UNETR(**(input_param))
        net.eval()
        with torch.no_grad():
            torch.jit.script(net)

        test_data = torch.randn(input_shape)
        test_script_save(net, test_data)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import numpy as np
from parameterized import parameterized

from monai.transforms import HistogramNormalize
from monai.utils import get_equivalent_dtype
from tests.test_utils import TEST_NDARRAYS, assert_allclose

TESTS = []
for p in TEST_NDARRAYS:
    TESTS.append(
        [
            {"num_bins": 4, "min": 1, "max": 5, "mask": np.array([1, 1, 1, 1, 1, 0])},
            p(np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0])),
            p(np.array([1.0, 1.5, 2.5, 4.0, 5.0, 5.0])),
        ]
    )

    TESTS.append(
        [
            {"num_bins": 4, "max": 4, "dtype": np.uint8},
            p(np.array([0.0, 1.0, 2.0, 3.0, 4.0])),
            p(np.array([0, 0, 1, 3, 4])),
        ]
    )

    TESTS.append(
        [
            {"num_bins": 256, "max": 255, "dtype": np.uint8},
            p(np.array([[[100.0, 200.0], [150.0, 250.0]]])),
            p(np.array([[[0, 170], [70, 255]]])),
        ]
    )


class TestHistogramNormalize(unittest.TestCase):
    @parameterized.expand(TESTS)
    def test_value(self, arguments, image, expected_data):
        result = HistogramNormalize(**arguments)(image)
        assert_allclose(result, expected_data, type_test="tensor")
        self.assertEqual(get_equivalent_dtype(result.dtype, data_type=np.ndarray), arguments.get("dtype", np.float32))


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os
import unittest
from pathlib import Path

import numpy as np
from parameterized import parameterized

from monai.transforms import SignalRandAddSinePartial
from monai.utils.type_conversion import convert_to_tensor

TESTS_PATH = Path(__file__).parents[1]
TEST_SIGNAL = os.path.join(TESTS_PATH, "testing_data", "signal.npy")
VALID_CASES = [([0.0, 1.0], [0.1, 0.6], [0.0, 0.4])]


class TestSignalRandAddSinePartialNumpy(unittest.TestCase):
    @parameterized.expand(VALID_CASES)
    def test_correct_parameters_multi_channels(self, boundaries, frequencies, fraction):
        self.assertIsInstance(SignalRandAddSinePartial(boundaries, frequencies, fraction), SignalRandAddSinePartial)
        sig = np.load(TEST_SIGNAL)
        partialsine = SignalRandAddSinePartial(boundaries, frequencies, fraction)
        partialsinesignal = partialsine(sig)
        self.assertEqual(partialsinesignal.shape[1], sig.shape[1])


class TestSignalRandAddSinePartialTorch(unittest.TestCase):
    @parameterized.expand(VALID_CASES)
    def test_correct_parameters_multi_channels(self, boundaries, frequencies, fraction):
        self.assertIsInstance(SignalRandAddSinePartial(boundaries, frequencies, fraction), SignalRandAddSinePartial)
        sig = convert_to_tensor(np.load(TEST_SIGNAL))
        partialsine = SignalRandAddSinePartial(boundaries, frequencies, fraction)
        partialsinesignal = partialsine(sig)
        self.assertEqual(partialsinesignal.shape[1], sig.shape[1])


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import numpy as np
import torch

from monai.utils import get_seed, set_determinism
from tests.test_utils import SkipIfBeforePyTorchVersion, skip_if_no_cuda


class TestSetDeterminism(unittest.TestCase):
    def test_values(self):
        # check system default flags
        set_determinism(None)
        self.assertTrue(not torch.backends.cudnn.deterministic)
        self.assertTrue(get_seed() is None)
        # set default seed
        set_determinism()
        self.assertTrue(get_seed() is not None)
        self.assertTrue(torch.backends.cudnn.deterministic)
        self.assertTrue(not torch.backends.cudnn.benchmark)
        # resume default
        set_determinism(None)
        self.assertTrue(not torch.backends.cudnn.deterministic)
        self.assertTrue(not torch.backends.cudnn.benchmark)
        self.assertTrue(get_seed() is None)
        # test seeds
        seed = 255
        set_determinism(seed=seed)
        self.assertEqual(seed, get_seed())
        a = np.random.randint(seed)
        b = torch.randint(seed, (1,))
        # tset when global flag support is disabled
        torch.backends.disable_global_flags()
        set_determinism(seed=seed)
        c = np.random.randint(seed)
        d = torch.randint(seed, (1,))
        self.assertEqual(a, c)
        self.assertEqual(b, d)
        self.assertTrue(torch.backends.cudnn.deterministic)
        self.assertTrue(not torch.backends.cudnn.benchmark)
        set_determinism(seed=None)


class TestSetFlag(unittest.TestCase):
    def setUp(self):
        set_determinism(1, use_deterministic_algorithms=True)

    @SkipIfBeforePyTorchVersion((1, 8))  # beta feature
    @skip_if_no_cuda
    def test_algo(self):
        with self.assertRaises(RuntimeError):
            x = torch.randn(20, 16, 50, 44, 31, requires_grad=True, device="cuda:0")
            y = torch.nn.AvgPool3d((3, 2, 2), stride=(2, 1, 2))(x)
            y.sum().backward()

    def tearDown(self):
        set_determinism(None, use_deterministic_algorithms=False)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest
from copy import deepcopy

import numpy as np
import torch
from parameterized import parameterized

from monai.data.meta_obj import set_track_meta
from monai.data.meta_tensor import MetaTensor
from monai.data.utils import to_affine_nd
from monai.transforms import SpatialResample
from monai.utils import optional_import
from tests.lazy_transforms_utils import test_resampler_lazy
from tests.test_utils import TEST_DEVICES, TEST_NDARRAYS_ALL, assert_allclose

TESTS = []

destinations_3d = [
    torch.tensor([[1.0, 0.0, 0.0, 0.0], [0.0, -1.0, 0.0, 1.0], [0.0, 0.0, 1.0, 0.0], [0.0, 0.0, 0.0, 1.0]]),
    torch.tensor([[-1.0, 0.0, 0.0, 1.0], [0.0, 1.0, 0.0, 0.0], [0.0, 0.0, 1.0, 0.0], [0.0, 0.0, 0.0, 1.0]]),
]
expected_3d = [
    torch.tensor([[[[4.0, 5.0, 6.0], [1.0, 2.0, 3.0]], [[10.0, 11.0, 12.0], [7.0, 8.0, 9.0]]]]),
    torch.tensor([[[[7.0, 8.0, 9.0], [10.0, 11.0, 12.0]], [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]]]),
]

for dst, expct in zip(destinations_3d, expected_3d):
    for device in TEST_DEVICES:
        for align in (False, True):
            interp = ("nearest", "bilinear")
            for interp_mode in interp:
                for padding_mode in ("zeros", "border", "reflection"):
                    TESTS.append(
                        [
                            torch.arange(12).reshape((1, 2, 2, 3)) + 1.0,  # data
                            *device,
                            {
                                "dst_affine": dst,
                                "dtype": torch.float64,
                                "align_corners": align,
                                "mode": interp_mode,
                                "padding_mode": padding_mode,
                            },
                            expct,
                        ]
                    )
if optional_import("cupy")[1] and optional_import("scipy.ndimage")[1]:
    TESTS.append(deepcopy(TESTS[-1]))
    TESTS[-1][2].update({"align_corners": True, "mode": 1, "padding_mode": "reflect"})  # type: ignore

destinations_2d = [
    torch.tensor([[1.0, 0.0, 0.0], [0.0, -1.0, 1.0], [0.0, 0.0, 1.0]]),  # flip the second
    torch.tensor([[-1.0, 0.0, 1.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]),  # flip the first
]
expected_2d = [torch.tensor([[[2.0, 1.0], [4.0, 3.0]]]), torch.tensor([[[3.0, 4.0], [1.0, 2.0]]])]

for dst, expct in zip(destinations_2d, expected_2d):
    for device in TEST_DEVICES:
        for align in (False, True):
            for interp_mode in ("nearest", "bilinear"):
                TESTS.append(
                    [
                        torch.arange(4).reshape((1, 2, 2)) + 1.0,
                        *device,
                        {
                            "dst_affine": dst,
                            "dtype": torch.float32,
                            "align_corners": align,
                            "mode": interp_mode,
                            "padding_mode": "zeros",
                        },
                        expct,
                    ]
                )

TEST_4_5_D = []
for device in TEST_DEVICES:
    for dtype in (torch.float32, torch.float64):
        # 4D
        TEST_4_5_D.append(
            [
                (1, 2, 2, 3, 1),
                (1, 1, 1, 1, 2),
                *device,
                dtype,
                torch.tensor(
                    [
                        [[[0.5, 0.0], [0.0, 2.0], [1.5, 1.0]], [[3.5, 3.0], [3.0, 5.0], [4.5, 4.0]]],
                        [[[6.5, 6.0], [6.0, 8.0], [7.5, 7.0]], [[9.5, 9.0], [9.0, 11.0], [10.5, 10.0]]],
                    ]
                ),
            ]
        )
        # 5D
        TEST_4_5_D.append(
            [
                (1, 2, 2, 3, 1, 1),
                (1, 1, 1, 1, 2, 2),
                *device,
                dtype,
                torch.tensor(
                    [
                        [
                            [[[0.0, 0.0], [0.0, 1.0]], [[0.5, 0.0], [1.5, 1.0]], [[1.0, 2.0], [2.0, 2.0]]],
                            [[[3.0, 3.0], [3.0, 4.0]], [[3.5, 3.0], [4.5, 4.0]], [[4.0, 5.0], [5.0, 5.0]]],
                        ],
                        [
                            [[[6.0, 6.0], [6.0, 7.0]], [[6.5, 6.0], [7.5, 7.0]], [[7.0, 8.0], [8.0, 8.0]]],
                            [[[9.0, 9.0], [9.0, 10.0]], [[9.5, 9.0], [10.5, 10.0]], [[10.0, 11.0], [11.0, 11.0]]],
                        ],
                    ]
                ),
            ]
        )

TEST_TORCH_INPUT = []
for track_meta in (True,):
    for t in TEST_4_5_D:
        TEST_TORCH_INPUT.append(t + [track_meta])


class TestSpatialResample(unittest.TestCase):
    @parameterized.expand(TESTS)
    def test_flips(self, img, device, data_param, expected_output):
        for p in TEST_NDARRAYS_ALL:
            img = p(img)
            if isinstance(img, MetaTensor):
                img.affine = torch.eye(4)
            if hasattr(img, "to"):
                img = img.to(device)
            resampler = SpatialResample()
            call_param = data_param.copy()
            call_param["img"] = img
            out = resampler(**call_param)
            assert_allclose(out, expected_output, rtol=1e-2, atol=1e-2)
            assert_allclose(to_affine_nd(len(out.shape) - 1, out.affine), call_param["dst_affine"])

            test_resampler_lazy(resampler, out, init_param=None, call_param=call_param)

    @parameterized.expand(TEST_4_5_D)
    def test_4d_5d(self, new_shape, tile, device, dtype, expected_data):
        img = np.arange(12).reshape(new_shape)
        img = np.tile(img, tile)
        img = MetaTensor(img).to(device)

        dst = torch.tensor([[1.0, 0.0, 0.0, 0.0], [0.0, 1.0, 0.0, 0.0], [0.0, 0.0, -1.0, 1.5], [0.0, 0.0, 0.0, 1.0]])
        dst = dst.to(dtype)
        init_param = {"dtype": dtype, "align_corners": True}
        call_param = {"img": img, "dst_affine": dst, "align_corners": False}
        resampler = SpatialResample(**init_param)
        out = resampler(**call_param)
        assert_allclose(out, expected_data[None], rtol=1e-2, atol=1e-2)
        assert_allclose(out.affine, dst.to(torch.float32), rtol=1e-2, atol=1e-2)

        test_resampler_lazy(resampler, out, init_param, call_param)

    @parameterized.expand(TEST_DEVICES)
    def test_ill_affine(self, device):
        img = MetaTensor(torch.arange(12).reshape(1, 2, 2, 3)).to(device)
        ill_affine = torch.tensor([[1, 0, 0, 0], [0, 0, 0, 0], [0, 0, -1, 1.5], [0, 0, 0, 1]])
        with self.assertRaises(ValueError):
            img.affine = torch.eye(4)
            dst_affine = ill_affine
            SpatialResample()(img=img, dst_affine=dst_affine)
        with self.assertRaises(ValueError):
            img.affine = ill_affine
            dst_affine = torch.eye(4)
            SpatialResample()(img=img, dst_affine=dst_affine)
        if not (optional_import("scipy")[1] and optional_import("cupy")[1]):
            return
        with self.assertRaises(ValueError):  # requires scipy
            SpatialResample(mode=1, align_corners=True)(img=img, dst_affine=dst_affine)
        with self.assertRaises(ValueError):
            SpatialResample(mode=1, align_corners=False)(img=img, dst_affine=dst_affine)

    @parameterized.expand(TEST_TORCH_INPUT)
    def test_input_torch(self, new_shape, tile, device, dtype, expected_data, track_meta):
        set_track_meta(track_meta)
        img = np.arange(12).reshape(new_shape)
        img = torch.as_tensor(np.tile(img, tile)).to(device)
        dst = torch.tensor([[1.0, 0.0, 0.0, 0.0], [0.0, 1.0, 0.0, 0.0], [0.0, 0.0, -1.0, 1.5], [0.0, 0.0, 0.0, 1.0]])
        dst = dst.to(dtype).to(device)
        init_param = {"dtype": dtype}
        call_param = {"img": img, "dst_affine": dst}
        resampler = SpatialResample(**init_param)
        out = resampler(**call_param)
        assert_allclose(out, expected_data[None], rtol=1e-2, atol=1e-2)

        test_resampler_lazy(resampler, out, init_param, call_param)

        if track_meta:
            self.assertIsInstance(out, MetaTensor)
            assert_allclose(out.affine, dst.to(torch.float32), rtol=1e-2, atol=1e-2)
        else:
            self.assertIsInstance(out, torch.Tensor)
            self.assertNotIsInstance(out, MetaTensor)

    @parameterized.expand(TESTS)
    def test_inverse(self, img, device, data_param, expected_output):
        img = MetaTensor(img, affine=torch.eye(4)).to(device)
        tr = SpatialResample()
        out = tr(img=img, **data_param)
        assert_allclose(out, expected_output, rtol=1e-2, atol=1e-2)
        assert_allclose(to_affine_nd(len(out.shape) - 1, out.affine), data_param["dst_affine"])

        # inverse
        out = tr.inverse(out)
        assert_allclose(img, out)
        expected_affine = to_affine_nd(len(out.affine) - 1, torch.eye(4))
        assert_allclose(out.affine, expected_affine)

    def test_unchange(self):
        for i, p in enumerate(TEST_NDARRAYS_ALL):
            set_track_meta(i % 2)
            img = p(np.arange(12).reshape(1, 3, 4))
            result = SpatialResample()(img)
            assert_allclose(result, img, type_test=False)
        set_track_meta(True)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import torch
from parameterized import parameterized

from monai.networks.nets import DenseNet, DenseNet121, SEResNet50
from monai.visualize import CAM

# 2D
TEST_CASE_0 = [
    {
        "model": "densenet2d",
        "shape": (2, 1, 48, 64),
        "feature_shape": (2, 1, 1, 2),
        "target_layers": "class_layers.relu",
        "fc_layers": "class_layers.out",
    },
    (2, 1, 48, 64),
]
# 3D
TEST_CASE_1 = [
    {
        "model": "densenet3d",
        "shape": (2, 1, 6, 6, 6),
        "feature_shape": (2, 1, 2, 2, 2),
        "target_layers": "class_layers.relu",
        "fc_layers": "class_layers.out",
    },
    (2, 1, 6, 6, 6),
]
# 2D
TEST_CASE_2 = [
    {
        "model": "senet2d",
        "shape": (2, 3, 64, 64),
        "feature_shape": (2, 1, 2, 2),
        "target_layers": "layer4",
        "fc_layers": "last_linear",
    },
    (2, 1, 64, 64),
]

# 3D
TEST_CASE_3 = [
    {
        "model": "senet3d",
        "shape": (2, 3, 8, 8, 48),
        "feature_shape": (2, 1, 1, 1, 2),
        "target_layers": "layer4",
        "fc_layers": "last_linear",
    },
    (2, 1, 8, 8, 48),
]


class TestClassActivationMap(unittest.TestCase):

    @parameterized.expand([TEST_CASE_0, TEST_CASE_1, TEST_CASE_2, TEST_CASE_3])
    def test_shape(self, input_data, expected_shape):
        model = None

        if input_data["model"] == "densenet2d":
            model = DenseNet121(spatial_dims=2, in_channels=1, out_channels=3)
        if input_data["model"] == "densenet3d":
            model = DenseNet(
                spatial_dims=3, in_channels=1, out_channels=3, init_features=2, growth_rate=2, block_config=(6,)
            )
        if input_data["model"] == "senet2d":
            model = SEResNet50(spatial_dims=2, in_channels=3, num_classes=4)
        if input_data["model"] == "senet3d":
            model = SEResNet50(spatial_dims=3, in_channels=3, num_classes=4)

        device = "cuda:0" if torch.cuda.is_available() else "cpu"
        model.to(device)
        model.eval()
        cam = CAM(nn_module=model, target_layers=input_data["target_layers"], fc_layers=input_data["fc_layers"])
        image = torch.rand(input_data["shape"], device=device)
        result = cam(x=image, layer_idx=-1)
        fea_shape = cam.feature_map_size(input_data["shape"], device=device)
        self.assertTupleEqual(fea_shape, input_data["feature_shape"])
        self.assertTupleEqual(result.shape, expected_shape)


if __name__ == "__main__":
    unittest.main()
